import http
from collections.abc import Callable
from functools import wraps
from ssl import SSLContext
from typing import Any, TypeVar

from adaptix import NameStyle, Retort, name_mapping
from aiohttp import ClientResponse, ClientSession, TCPConnector
from dataclass_rest import get
from dataclass_rest.client_protocol import FactoryProtocol
from dataclass_rest.http.aiohttp import AiohttpClient, AiohttpMethod
from dataclass_rest.http_request import HttpRequest

from .models import PagingResponse, Status

Func = TypeVar("Func", bound=Callable)


def _collect_by_pages(func: Func) -> Func:
    """Collect all results using only pagination."""

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        kwargs.setdefault("offset", 0)
        limit = kwargs.setdefault("limit", 100)
        results = []
        method = func.__get__(self, self.__class__)
        has_next = True
        while has_next:
            page = await method(*args, **kwargs)
            kwargs["offset"] += limit
            results.extend(page.results)
            has_next = bool(page.next)
        return PagingResponse(
            previous=None,
            next=None,
            count=len(results),
            results=results,
        )

    return wrapper


# default batch size 100 is calculated to fit list of UUIDs in 4k URL length
def collect(func: Func, field: str = "", batch_size: int = 100) -> Func:
    """
    Collect data from method iterating over pages and filter batches.

    :param func: Method to call
    :param field: Field which defines a filter split into batches
    :param batch_size: Limit of values in `field` filter requested at a time
    """
    func = _collect_by_pages(func)
    if not field:
        return func

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        method = func.__get__(self, self.__class__)

        value = kwargs.get(field)
        if not value:
            return await method(*args, **kwargs)

        results = []
        for offset in range(0, len(value), batch_size):
            kwargs[field] = value[offset : offset + batch_size]
            page = await method(*args, **kwargs)
            results.extend(page.results)
        return PagingResponse(
            previous=None,
            next=None,
            count=len(results),
            results=results,
        )

    return wrapper


class NoneAwareAiohttpMethod(AiohttpMethod):
    async def _pre_process_request(self, request: HttpRequest) -> HttpRequest:
        request.query_params = {
            k: v for k, v in request.query_params.items() if v is not None
        }
        return request

    async def _response_body(self, response: ClientResponse) -> Any:
        if response.status == http.HTTPStatus.NO_CONTENT:
            return None
        return await super()._response_body(response)


class BaseNetboxClient(AiohttpClient):
    method_class = NoneAwareAiohttpMethod

    def __init__(
        self,
        url: str,
        token: str = "",
        ssl_context: SSLContext | None = None,
    ):
        url = url.rstrip("/") + "/api/"

        connector = TCPConnector(ssl=ssl_context)
        session = ClientSession(connector=connector)
        if token:
            session.headers["Authorization"] = f"Token {token}"
        super().__init__(url, session)

    async def close(self):
        await self.session.close()


class NetboxStatusClient(BaseNetboxClient):
    def _init_response_body_factory(self) -> FactoryProtocol:
        return Retort(recipe=[name_mapping(name_style=NameStyle.LOWER_KEBAB)])

    @get("status")
    async def status(self) -> Status: ...



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any


@dataclass
class Entity:
    id: int
    name: str
    display: str
    url: str


@dataclass
class Label:
    value: str
    label: str


@dataclass
class DeviceType:
    id: int
    manufacturer: Entity
    model: str


@dataclass
class DeviceIp:
    id: int
    display: str
    address: str
    family: int


@dataclass
class LinkPeer(Entity):
    cable: int
    device: Entity


@dataclass
class InterfaceCable:
    id: int
    label: str
    display: str
    url: str


@dataclass
class InterfaceType:
    value: str
    label: str


@dataclass
class InterfaceConnectedEndpoint(Entity):
    device: Entity


@dataclass
class Interface(Entity):
    cable: InterfaceCable | None
    cable_end: str
    device: Entity
    link_peers: list[LinkPeer]
    link_peers_type: str | None
    enabled: bool
    type: InterfaceType
    description: str
    connected_endpoints: list[InterfaceConnectedEndpoint] | None
    created: datetime
    last_updated: datetime


@dataclass
class Device(Entity):
    url: str
    display: str  # renamed in 3.x from display_name
    device_type: DeviceType
    device_role: Entity
    tenant: Entity | None
    platform: Entity | None
    serial: str
    asset_tag: str | None
    site: Entity
    rack: Entity | None
    position: float | None
    face: Label | None
    status: Label
    primary_ip: DeviceIp | None
    primary_ip4: DeviceIp | None
    primary_ip6: DeviceIp | None
    tags: list[Entity]
    custom_fields: dict[str, Any]
    created: datetime
    last_updated: datetime


@dataclass
class IpFamily:
    value: int
    label: str


@dataclass
class IpAddress:
    id: int
    assigned_object_id: int | None
    display: str
    family: IpFamily
    address: str
    status: Label
    tags: list[Entity]
    created: datetime
    last_updated: datetime


class CableType(str, Enum):
    AOC = "aoc"
    CAT3 = "cat3"
    CAT5 = "cat5"
    CAT5E = "cat5e"
    CAT6 = "cat6"
    CAT6A = "cat6a"
    CAT7 = "cat7"
    CAT7A = "cat7a"
    CAT8 = "cat8"
    COAXIAL = "coaxial"
    DAC_ACTIVE = "dac-active"
    DAC_PASSIVE = "dac-passive"
    MMF = "mmf"
    MMF_OM1 = "mmf-om1"
    MMF_OM2 = "mmf-om2"
    MMF_OM3 = "mmf-om3"
    MMF_OM4 = "mmf-om4"
    MMF_OM5 = "mmf-om5"
    MRJ21_TRUNK = "mrj21-trunk"
    POWER = "power"
    SMF = "smf"
    SMF_OS1 = "smf-os1"
    SMF_OS2 = "smf-os2"
    VALUE_23 = ""


@dataclass
class GenericObject:
    object_type: str
    object_id: int


@dataclass
class Cable:
    id: int
    url: str
    display: str
    created: None | datetime
    last_updated: None | datetime
    custom_fields: dict[str, Any]
    tags: list[Entity]
    type: CableType | None = None
    a_terminations: None | list[GenericObject] = None
    b_terminations: None | list[GenericObject] = None
    status: None | Label = None
    tenant: Entity | None = None
    label: None | str = None
    color: None | str = None
    length: None | float = None
    length_unit: Label | None = None
    description: None | str = None
    comments: None | str = None


@dataclass
class NewCable:
    custom_fields: dict[str, Any] | None = None
    tags: list[Entity] | None = None
    type: CableType | None = None
    a_terminations: None | list[GenericObject] = None
    b_terminations: None | list[GenericObject] = None
    status: None | Label = None
    tenant: Entity | None = None
    label: None | str = None
    color: None | str = None
    length: None | float = None
    length_unit: Label | None = None
    description: None | str = None
    comments: None | str = None


@dataclass
class ItemToDelete:
    id: int


@dataclass
class Prefix:
    id: int
    prefix: str
    site: Entity | None
    vrf: Entity | None
    tenant: Entity | None
    vlan: Entity | None
    role: Entity | None
    status: Label
    is_pool: bool
    custom_fields: dict[str, Any]
    created: datetime
    last_updated: datetime



from dataclasses import dataclass
from datetime import datetime
from typing import Any


@dataclass
class Entity:
    id: int
    name: str


@dataclass
class Label:
    value: int
    label: str


@dataclass
class DeviceType:
    id: int
    manufacturer: Entity
    model: str


@dataclass
class DeviceIp:
    id: int
    address: str
    family: int


@dataclass
class Device(Entity):
    url: str
    display_name: str
    device_type: DeviceType
    device_role: Entity
    tenant: Entity | None
    platform: Entity | None
    serial: str
    asset_tag: str | None
    site: Entity
    rack: Entity | None
    position: float | None
    face: Label | None
    status: Label
    primary_ip: DeviceIp | None
    primary_ip4: DeviceIp | None
    primary_ip6: DeviceIp | None
    tags: list[str]
    custom_fields: dict[str, Any]
    created: datetime
    last_updated: datetime


@dataclass
class Interface(Entity):
    device: Entity
    enabled: bool


@dataclass
class Vrf(Entity):
    rd: str


@dataclass
class IpAddress:
    id: int
    family: int
    address: str
    vrf: Vrf | None
    tenant: Any  # ???
    status: Label
    description: str | None
    custom_fields: dict[str, Any]
    tags: list[str]
    created: datetime
    last_updated: datetime

    interface: Entity

    nat_inside: Any  # ???
    nat_outside: Any  # ???



from datetime import datetime

import dateutil.parser
from adaptix import Retort, loader
from dataclass_rest import get

from annetbox.base.client_sync import BaseNetboxClient, collect
from annetbox.base.models import PagingResponse
from .models import Device, Interface, IpAddress


class NetboxV24(BaseNetboxClient):
    def _init_response_body_factory(self) -> Retort:
        return Retort(recipe=[loader(datetime, dateutil.parser.parse)])

    # dcim
    @get("dcim/interfaces/")
    def dcim_interfaces(
        self,
        device_id: list[int] | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> PagingResponse[Interface]:
        pass

    dcim_all_interfaces = collect(dcim_interfaces, field="device_id")

    @get("dcim/devices/")
    def dcim_devices(
        self,
        name: list[str] | None = None,
        tag: str | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> PagingResponse[Device]:
        pass

    dcim_all_devices = collect(dcim_devices)

    @get("dcim/devices/{device_id}")
    def dcim_device(
        self,
        device_id: int,
    ) -> Device:
        pass

    # ipam
    @get("ipam/ip-addresses/")
    def ipam_ip_addresses(
        self,
        interface_id: list[int] | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> PagingResponse[IpAddress]:
        pass

    ipam_all_ip_addresses = collect(ipam_ip_addresses, field="interface_id")



from dataclasses import dataclass
from typing import Generic, TypeVar

Model = TypeVar("Model")


@dataclass
class Status:
    netbox_version: str
    plugins: dict[str, str]


@dataclass
class PagingResponse(Generic[Model]):
    next: str | None
    previous: str | None
    count: int
    results: list[Model]



from collections.abc import Iterable
from datetime import datetime

import dateutil.parser
from adaptix import Retort, loader, name_mapping
from dataclass_rest import delete, get, post
from dataclass_rest.client_protocol import FactoryProtocol

from annetbox.base.client_sync import BaseNetboxClient, collect
from annetbox.base.models import PagingResponse
from .models import (
    Cable,
    Device,
    Interface,
    IpAddress,
    ItemToDelete,
    NewCable,
    Prefix,
)


class NetboxV37(BaseNetboxClient):
    def _init_response_body_factory(self) -> FactoryProtocol:
        return Retort(recipe=[loader(datetime, dateutil.parser.parse)])

    def _init_request_body_factory(self) -> FactoryProtocol:
        return Retort(
            recipe=[
                name_mapping(NewCable, omit_default=True),
            ],
        )

    # dcim
    @get("dcim/interfaces/")
    def dcim_interfaces(
        self,
        id: list[int] | None = None,
        device: list[str] | None = None,
        device__n: list[str] | None = None,
        device_id: list[int] | None = None,
        device_id__n: list[int] | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> PagingResponse[Interface]:
        pass

    dcim_all_interfaces = collect(dcim_interfaces, field="device_id")
    dcim_all_interfaces_by_id = collect(dcim_interfaces, field="id")

    @get("dcim/interfaces/{id}/")
    def dcim_interface(self, id: int) -> Interface:
        pass

    @get("dcim/cables/")
    def dcim_cables(
        self,
        device: list[str] | None = None,
        device_id: list[int] | None = None,
        interface_id: list[int] | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> PagingResponse[Cable]:
        pass

    dcim_all_cables = collect(dcim_cables, field="interface_id")

    @post("dcim/cables/")
    def dcim_cable_create(self, body: NewCable) -> Cable:
        pass

    @post("dcim/cables/")
    def dcim_cable_bulk_create(
        self,
        body: list[NewCable],
    ) -> list[Cable]:
        pass

    @delete("dcim/cables/")
    def _dcim_cable_bulk_delete(self, body: list[ItemToDelete]) -> None:
        pass

    def dcim_cable_bulk_delete(self, body: Iterable[int]) -> None:
        return self._dcim_cable_bulk_delete([ItemToDelete(id=x) for x in body])

    @delete("dcim/cables/{id}/")
    def dcim_cable_delete(self, id: int) -> None:
        pass

    @get("dcim/devices/")
    def dcim_devices(
        self,
        name: list[str] | None = None,
        name__empty: bool | None = None,
        name__ic: list[str] | None = None,
        name__ie: list[str] | None = None,
        name__iew: list[str] | None = None,
        name__isw: list[str] | None = None,
        name__n: list[str] | None = None,
        name__nic: list[str] | None = None,
        name__nie: list[str] | None = None,
        name__niew: list[str] | None = None,
        name__nisw: list[str] | None = None,
        id: list[int] | None = None,
        tag: list[str] | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> PagingResponse[Device]:
        pass

    dcim_all_devices = collect(dcim_devices)
    dcim_all_devices_by_id = collect(dcim_devices, field="id")

    @get("dcim/devices/{device_id}/")
    def dcim_device(
        self,
        device_id: int,
    ) -> Device:
        pass

    # ipam
    @get("ipam/ip-addresses/")
    def ipam_ip_addresses(
        self,
        interface_id: list[int] | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> PagingResponse[IpAddress]:
        pass

    ipam_all_ip_addresses = collect(ipam_ip_addresses, field="interface_id")

    @get("ipam/prefixes/")
    def prefixes(
        self,
        prefix: list[str] | None = None,
        limit: int = 20,
        offset: int = 0,
    ) -> PagingResponse[Prefix]:
        pass

    ipam_all_prefixes = collect(prefixes, field="prefix")
